module Views.LinePlot exposing (view, init, update)
import Model exposing (Global, View(..))
import Msg exposing (Msg(..))
import Data.DataLoader exposing (Entry)

import Html exposing (Html, div, select, option, label)
import Html.Attributes exposing (type_, style)
import Html.Events exposing (onClick, onInput)

import List.Extra exposing (unique)
import Dict exposing (Dict)

import Chart.Line as Line
import Scale.Color
import Chart.Symbol as Symbol exposing (Symbol)
import Axis
import Chart.Annotation as Annotation
import TypedSvg.Core exposing (Svg)

import Config.LinePlotConfig as Local exposing (Model)
import Html.Attributes exposing (wrap)

import Utils exposing (codeToName)

init : Model
init = { pointAnnotation = Nothing , vLineAnnotation = Nothing }

propagate : Local.Msg -> Msg
propagate msg =
    GotLinePlot msg

update : Local.Msg -> Global -> Model -> ( Global, Model, Cmd Msg )
update msg global model = 
    case msg of
        Local.CountryChange country ->
            ( { global | selectedCountry = Just country }, model, Cmd.none )
        Local.Hint response ->
            ( global, { model
                | pointAnnotation =
                    response
                        |> Maybe.map (\hint -> ( hint, [( "fill","#fff" ), ( "stroke","#000" ), ( "stroke-width","1.5" )] ))
                , vLineAnnotation =
                    response
                        |> Maybe.map (\hint -> ( hint, [( "stroke", "#999999" )] ))
              }
            , Cmd.none
            )



type alias DataPoint =
    { groupLabel : String
    , x : Int
    , y : Int
    }



view : Global -> Model -> Html Msg
view global model = 
    let
        -- Default selected country
        country_default : String
        country_default =
            Maybe.withDefault "AT" global.selectedCountry

        -- Get a list of all countries from global.data
        countries : List String
        countries =
            List.map .state global.data
            |> unique

        -- Generate a list of options for the select element
        country_option : String -> Html Msg
        country_option country =
            let
                selected : Bool
                selected =
                    case global.selectedCountry of
                        Nothing -> False
                        Just c -> country == c
            in
            option
            [ Html.Attributes.value country, Html.Attributes.selected selected ]
            [ Html.text (codeToName country) ]


        -- Filter global.data based on the selected visa type and country
        filterList : List Entry -> String -> String -> List (Int, Int)
        filterList data country visatype =
            data
            |> List.filter (\entry -> entry.state == country)
            |> List.map (\entry -> (entry.year, Maybe.withDefault 0 (
                case visatype of
                    "totalVisasApplied" -> entry.totalVisasApplied
                    "totalVisasIssued" -> entry.totalVisasIssued
                    _ -> Nothing
                )))
            -- Merge tuples of the same year (sum the visa count)
            |> List.foldl
                    (\(year, count) acc ->
                        Dict.update year
                            (\maybeCount ->
                                Just (Maybe.withDefault 0 maybeCount + count)
                            )
                            acc
                    )
                    Dict.empty
            |> Dict.toList
            |> List.map (\(year, count) -> (year, round count))


        -- Get a list of all years from global.data for the xaxis labeling of the selected country
        getYears : List Int
        getYears =
            let
                applied_years : List Int
                applied_years = List.map Tuple.first (filterList global.data country_default "totalVisasApplied")

                issued_years : List Int
                issued_years = List.map Tuple.first (filterList global.data country_default "totalVisasIssued")
            in
            List.append applied_years issued_years
                |> List.sort
                |> unique

        
        -- Wrap the result from filterList in List DataPoint
        wrapDataPoint : List (Int, Int) -> String -> List DataPoint
        wrapDataPoint data label =
            List.map (\(year, count) -> { groupLabel = label, x = year, y = count }) data

        
        filterVApplied : List DataPoint
        filterVApplied = wrapDataPoint (filterList global.data country_default "totalVisasApplied") "applied"

        filterVIssued : List DataPoint
        filterVIssued = wrapDataPoint (filterList global.data country_default "totalVisasIssued") "issued"

        -- Merge the two lists
        dataLines : List DataPoint
        dataLines = filterVApplied ++ filterVIssued


        accessor : Line.Accessor DataPoint
        accessor =
            Line.cont
                { xGroup = .groupLabel >> Just
                , xValue = .x >> toFloat
                , yValue = .y >> toFloat
                }

        sharedAttributes : List (Axis.Attribute value)
        sharedAttributes =
            [ Axis.tickSizeOuter 0
            , Axis.tickSizeInner 3
            ]


        yAxis : Line.YAxis Float
        yAxis =
            Line.axisLeft (Axis.tickCount 6 :: sharedAttributes)


        xAxisTicks : List Float
        xAxisTicks =
            getYears
                |> List.map toFloat


        xAxis : Line.XAxis Float
        xAxis =
            Line.axisBottom
                ([ Axis.ticks xAxisTicks
                , Axis.tickFormat (round >> String.fromInt)
                ]
                    ++ sharedAttributes
                )


        tooltip : Html Msg
        tooltip =
            let
                margin = { top = 10, right = 80, bottom = 20, left = 30 }

                height = 600

                xOffset =
                    margin.left + 90

                annotation =
                    model.pointAnnotation

                bottom =
                    annotation
                        |> Maybe.map Tuple.first
                        |> Maybe.map
                            (\{ yPosition } -> String.fromFloat (height - yPosition) ++ "px")
                            -- (\{ yPosition } -> String.fromFloat (height - 40) ++ "px")
                        |> Maybe.withDefault ""

                left =
                    annotation
                        |> Maybe.map Tuple.first
                        |> Maybe.map
                            (\{ xPosition } -> String.fromFloat (xPosition + xOffset) ++ "px")
                            -- (\{ xPosition } -> String.fromFloat (width - 20) ++ "px")
                        |> Maybe.withDefault ""

                display =
                    annotation
                        |> Maybe.map (always "block")
                        |> Maybe.withDefault "none"

                year =
                    annotation
                        |> Maybe.map
                            (Tuple.first
                                >> .selection
                                >> .x
                                >> String.fromFloat
                            )
                        |> Maybe.withDefault ""

                values =
                    annotation
                        |> Maybe.map
                            (Tuple.first
                                >> .selection
                                >> .y
                            )
                        |> Maybe.withDefault []
                        |> List.map
                            (\{ groupLabel, value } ->
                                Html.div []
                                    [ Html.text
                                        (Maybe.withDefault "" groupLabel
                                            ++ ": "
                                            ++ String.fromFloat value
                                        )
                                    ]
                            )
            in
            Html.div
                [ Html.Attributes.style "margin" "0"
                , Html.Attributes.style "padding" "5px"
                , Html.Attributes.style "font-size" "14px"
                , Html.Attributes.style "border" "1px #aaa solid"
                , Html.Attributes.style "position" "absolute"
                , Html.Attributes.style "bottom" bottom
                , Html.Attributes.style "left" left
                , Html.Attributes.style "background" "white"
                , Html.Attributes.style "opacity" "90%"
                , Html.Attributes.style "width" "130px"
                , Html.Attributes.style "display" display
                ]
                [ Html.text year
                , Html.br [] []
                , Html.div [] values
                ]


        -- -- Create the Line Plot
        chart : List DataPoint -> Html Msg
        chart data =
            let
                linit = Line.init
                    { margin = { top = 10, right = 80, bottom = 40, left = 60 }
                    , width = 800
                    , height = 600
                    }
                    |> Line.withColorPalette Scale.Color.tableau10
                    |> Line.withEvent (Line.hoverAll Local.Hint)
                    |> Line.withPointAnnotation model.pointAnnotation
                    |> Line.withVLineAnnotation model.vLineAnnotation
                    -- |> Line.withLabels Line.xGroupLabel
                    |> Line.withSymbols [ 
                        Symbol.circle
                        |> Symbol.withStyle [ ( "stroke", "white" ) ]
                        |> Symbol.withSize 8 ]
                    |> Line.withLineStyle [ ( "stroke-width", "2.5" ) ]
                    |> Line.withYAxis yAxis
                    |> Line.withXAxisCont xAxis
                    |> Line.render ( data, accessor )

            in
            Html.div [ Html.Attributes.style "position" "relative"
                     ]
                [ Html.map propagate linit
                , tooltip
                ]



        -- Handle country change
        handleCountryChange : String -> Msg
        handleCountryChange msg =
            msg
            |> Local.CountryChange
            |> propagate

    in
    div [  Html.Attributes.class "barplot-wrapper"  ]
        [ div [ Html.Attributes.style "display" "flex"
            , Html.Attributes.style "flex-direction" "column"
            , Html.Attributes.style "gap" "4px"
        ] [
            label [Html.Attributes.for "select_country"] [ Html.text "Choose Schengen State: " ]
            , select
                [ Html.Attributes.id "select_country"
                , Html.Attributes.name "country"
                , onInput handleCountryChange
                ] 
                    (List.map country_option countries)
        ] 
        , chart dataLines
        , div [ Html.Attributes.style "display" "flex"
                , Html.Attributes.style "flex-direction" "row"
                , Html.Attributes.style "gap" "4px"
                , Html.Attributes.style "margin" "auto"
        ] [
            div [ Html.Attributes.style "height" "15px"
                 , Html.Attributes.style "width" "15px"
                 , Html.Attributes.style "background-color" "#4e79a7"
                 , Html.Attributes.style "border-radius"" 50%"
                 , Html.Attributes.style "display" "inline-block"] []
            , div [] [ Html.text "applied" ]
            , div [ Html.Attributes.style "height" "15px"
                 , Html.Attributes.style "width" "15px"
                 , Html.Attributes.style "background-color" "#f28e2c"
                 , Html.Attributes.style "border-radius"" 50%"
                 , Html.Attributes.style "display" "inline-block"] []
            , div [] [ Html.text "issued" ]
        ]
        ]